/* Programme permettant NBECHANGES d'une structure de type structure_t
   entre un pere et un enfant via une zone de memoire partagee de 
   NBCASES de sizeof(structure_t) octets controlee a l'aide d'un
   mecanisme de producteur/consommateur
*/

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/wait.h>
#include <stdio.h>
#include <assert.h>
#include "declarations.h"

/* Memoire partagee */
structure_t *tampon;

/* Semaphores */
int infoPrete;
int placeDispo;

/* Procdures P et V pour manipulation du smaphore #0 des objets 
   smaphores                                                     
*/
void P(int semId){
  struct sembuf aBuf = {0, -1, 0};
  int retour;
  retour = semop(semId, &aBuf, 1);
  assert(retour != -1);
}

void V(int semId){
  struct sembuf aBuf = {0, +1, 0};
  int retour;
  retour = semop(semId, &aBuf, 1);
  assert(retour != -1);
}



/* Pere et enfant
 */
void pere(){
  int i;
  int iDepot = 0;
  int status;
  structure_t s;

  /* Initialisation de tous les octets de s  \0 */
  memset(&s, 0, sizeof(s));

  /* On envoie s NBECHANGES fois a l'enfant via le tampon */
  for (i=0 ; i<NBECHANGES ; i++){
    /* Envoi de s dans le tampon */
    P(placeDispo);
    tampon[iDepot] = s;
    iDepot = (iDepot + 1) % NBCASES;
    V(infoPrete);
  }

  /* Attente deces de l'enfant */
  wait(&status);
}

void enfant(){
  int i;
  int iExtrait = 0;
  structure_t s;

  /* On lit NBECHANGES fois dans le tamponla structure s envoyee par */
  /* le pere                                                         */
  for (i=0 ; i<NBECHANGES ; i++){
    /* Lecture de s dans le tampon */
    P(infoPrete);
    s = tampon[iExtrait];
    iExtrait = (iExtrait + 1) % NBCASES;
    V(placeDispo);
  }
}

/* main
 */
int main(){
  int shmid;
  int retour;

  /* Creation de la zone de memoire partagee correspondant au tampon */
  shmid = shmget(IPC_PRIVATE,
		 NBCASES*sizeof(structure_t),
		 0600|IPC_CREAT);
  assert( shmid != -1);
  tampon = (structure_t*) shmat(shmid, NULL, 0);
  assert(tampon != (structure_t*)-1);

  /* Creation des semaphores */
  infoPrete = semget(IPC_PRIVATE, 1, 0600|IPC_CREAT);
  assert(infoPrete != -1);

  placeDispo = semget(IPC_PRIVATE, 1, 0600|IPC_CREAT);
  assert(placeDispo != -1);

  /* Initialisation des differents semaphores */
  retour = semctl(infoPrete, 0, SETVAL, 0);
  assert(retour != -1);

  retour = semctl(placeDispo, 0, SETVAL, NBCASES);
  assert(retour != -1);

  /* Creation de l'enfant */
  switch(fork()){
  case -1:
    perror("fork");
    return EXIT_FAILURE;
  case 0:
    enfant();
    /* On se detache du tampon puisqu'on ne s'en sert plus */
    retour = shmdt(tampon);
    assert(retour == 0);
    break;
  default:
    pere();
    /* On se detache du tampon puisqu'on ne s'en sert plus */
    retour = shmdt(tampon);
    assert(retour == 0);
    /* On supprime cette zone de memoire partagee */
    retour = shmctl(shmid,IPC_RMID, NULL);
    assert(retour == 0);
    /* On supprime les semaphores */
    retour = semctl(infoPrete,0,IPC_RMID, 0);
    assert(retour == 0);
    retour = semctl(placeDispo,0,IPC_RMID, 0);
    assert(retour == 0);
    break;
  }
  return EXIT_SUCCESS;
}
